'use strict';

const node_path = require('node:path');
const fs = require('fs-extra');
const os = require('os');
const kolorist = require('kolorist');
const glob = require('fast-glob');
const debug = require('debug');
const tsMorph = require('ts-morph');
const vite = require('vite');
const typescript = require('typescript');
const pluginutils = require('@rollup/pluginutils');
const node_fs = require('node:fs');
const node_module = require('node:module');
const apiExtractor = require('@microsoft/api-extractor');
const Collector_js = require('@microsoft/api-extractor/lib/collector/Collector.js');
const MessageRouter_js = require('@microsoft/api-extractor/lib/collector/MessageRouter.js');
const SourceMapper_js = require('@microsoft/api-extractor/lib/collector/SourceMapper.js');
const DtsRollupGenerator_js = require('@microsoft/api-extractor/lib/generators/DtsRollupGenerator.js');
const nodeCoreLibrary = require('@rushstack/node-core-library');

function isNativeObj(value) {
  return Object.prototype.toString.call(value) === "[object Object]";
}
function isRegExp(value) {
  return Object.prototype.toString.call(value) === "[object RegExp]";
}
function isPromise(value) {
  return !!value && typeof value.then === "function" && typeof value.catch === "function";
}
function mergeObjects(sourceObj, targetObj) {
  const loop = [
    {
      source: sourceObj,
      target: targetObj
    }
  ];
  while (loop.length) {
    const { source, target } = loop.pop();
    Object.keys(target).forEach((key) => {
      if (isNativeObj(target[key])) {
        if (!isNativeObj(source[key])) {
          source[key] = {};
        }
        loop.push({
          source: source[key],
          target: target[key]
        });
      } else if (Array.isArray(target[key])) {
        if (!Array.isArray(source[key])) {
          source[key] = [];
        }
        loop.push({
          source: source[key],
          target: target[key]
        });
      } else {
        source[key] = target[key];
      }
    });
  }
  return sourceObj;
}
function ensureAbsolute(path, root) {
  return path ? node_path.isAbsolute(path) ? path : node_path.resolve(root, path) : root;
}
function ensureArray(value) {
  return Array.isArray(value) ? value : value ? [value] : [];
}
async function runParallel(maxConcurrency, source, iteratorFn) {
  const ret = [];
  const executing = [];
  for (const item of source) {
    const p = Promise.resolve().then(() => iteratorFn(item, source));
    ret.push(p);
    if (maxConcurrency <= source.length) {
      const e = p.then(() => executing.splice(executing.indexOf(e), 1));
      executing.push(e);
      if (executing.length >= maxConcurrency) {
        await Promise.race(executing);
      }
    }
  }
  return Promise.all(ret);
}
const speRE = /[\\/]/;
function queryPublicPath(paths) {
  if (paths.length === 0) {
    return "";
  } else if (paths.length === 1) {
    return node_path.dirname(paths[0]);
  }
  let publicPath = node_path.normalize(node_path.dirname(paths[0])) + node_path.sep;
  let publicUnits = publicPath.split(speRE);
  let index = publicUnits.length - 1;
  for (const path of paths.slice(1)) {
    if (!index) {
      return publicPath;
    }
    const dirPath = node_path.normalize(node_path.dirname(path)) + node_path.sep;
    if (dirPath.startsWith(publicPath)) {
      continue;
    }
    const units = dirPath.split(speRE);
    if (units.length < index) {
      publicPath = dirPath;
      publicUnits = units;
      continue;
    }
    for (let i = 0; i <= index; ++i) {
      if (publicUnits[i] !== units[i]) {
        if (!i) {
          return "";
        }
        index = i - 1;
        publicUnits = publicUnits.slice(0, index + 1);
        publicPath = publicUnits.join(node_path.sep) + node_path.sep;
        break;
      }
    }
  }
  return publicPath.slice(0, -1);
}
function removeDirIfEmpty(dir) {
  if (!node_fs.existsSync(dir)) {
    return;
  }
  let onlyHasDir = true;
  for (const file of node_fs.readdirSync(dir)) {
    const abs = node_path.resolve(dir, file);
    if (node_fs.lstatSync(abs).isDirectory()) {
      if (!removeDirIfEmpty(abs)) {
        onlyHasDir = false;
      }
    } else {
      onlyHasDir = false;
    }
  }
  if (onlyHasDir) {
    node_fs.rmdirSync(dir);
  }
  return onlyHasDir;
}

const globSuffixRE = /^((?:.*\.[^.]+)|(?:\*+))$/;
function normalizeGlob(path) {
  if (/[\\/]$/.test(path)) {
    return path + "**";
  } else if (!globSuffixRE.test(path.split(/[\\/]/).pop())) {
    return path + "/**";
  }
  return path;
}
const globalDynamicTypeRE = /import\(['"][^;\n]+?['"]\)\.\w+[.()[\]<>,;\n\s]/g;
const dynamicTypeRE = /import\(['"](.+)['"]\)\.(.+)([.()[\]<>,;\n\s])/;
const importTypesRE = /import\s?(?:type)?\s?\{(.+)\}\s?from\s?['"].+['"]/;
function transformDynamicImport(content) {
  const importMap = /* @__PURE__ */ new Map();
  content = content.replace(globalDynamicTypeRE, (str) => {
    const matchResult = str.match(dynamicTypeRE);
    const libName = matchResult[1];
    const importSet = importMap.get(libName) ?? importMap.set(libName, /* @__PURE__ */ new Set()).get(libName);
    const usedType = matchResult[2];
    importSet.add(usedType);
    return usedType + matchResult[3];
  });
  importMap.forEach((importSet, libName) => {
    const importReg = new RegExp(
      `import\\s?(?:type)?\\s?\\{[^;\\n]+\\}\\s?from\\s?['"]${libName}['"]`,
      "g"
    );
    const matchResult = content.match(importReg);
    if (matchResult?.[0]) {
      const importedTypes = matchResult[0].match(importTypesRE)[1].trim().split(",");
      content = content.replace(
        matchResult[0],
        `import type { ${Array.from(importSet).concat(importedTypes).join(", ")} } from '${libName}'`
      );
    } else {
      content = `import type { ${Array.from(importSet).join(", ")} } from '${libName}';
` + content;
    }
  });
  return content;
}
function isAliasMatch(alias, importee) {
  if (isRegExp(alias.find))
    return alias.find.test(importee);
  if (importee.length < alias.find.length)
    return false;
  if (importee === alias.find)
    return true;
  return importee.indexOf(alias.find) === 0 && (alias.find.endsWith("/") || importee.substring(alias.find.length)[0] === "/");
}
const globalImportRE = /(?:(?:import|export)\s?(?:type)?\s?(?:(?:\{[^;\n]+\})|(?:[^;\n]+))\s?from\s?['"][^;\n]+['"])|(?:import\(['"][^;\n]+?['"]\))/g;
const staticImportRE = /(?:import|export)\s?(?:type)?\s?\{?.+\}?\s?from\s?['"](.+)['"]/;
const dynamicImportRE = /import\(['"]([^;\n]+?)['"]\)/;
const simpleStaticImportRE = /((?:import|export).+from\s?)['"](.+)['"]/;
const simpleDynamicImportRE = /(import\()['"](.+)['"]\)/;
function transformAliasImport(filePath, content, aliases, exclude = []) {
  if (!aliases.length)
    return content;
  return content.replace(globalImportRE, (str) => {
    let matchResult = str.match(staticImportRE);
    let isDynamic = false;
    if (!matchResult) {
      matchResult = str.match(dynamicImportRE);
      isDynamic = true;
    }
    if (matchResult?.[1]) {
      const matchedAlias = aliases.find((alias) => isAliasMatch(alias, matchResult[1]));
      if (matchedAlias) {
        if (exclude.some((e) => isRegExp(e) ? e.test(matchResult[1]) : String(e) === matchResult[1])) {
          return str;
        }
        const truthPath = node_path.isAbsolute(matchedAlias.replacement) ? vite.normalizePath(node_path.relative(node_path.dirname(filePath), matchedAlias.replacement)) : vite.normalizePath(matchedAlias.replacement);
        return str.replace(
          isDynamic ? simpleDynamicImportRE : simpleStaticImportRE,
          `$1'${matchResult[1].replace(
            matchedAlias.find,
            (truthPath.startsWith(".") ? truthPath : `./${truthPath}`) + (typeof matchedAlias.find === "string" && matchedAlias.find.endsWith("/") ? "/" : "")
          )}'${isDynamic ? ")" : ""}`
        );
      }
    }
    return str;
  });
}
const pureImportRE = /import\s?['"][^;\n]+?['"];?\n?/g;
function removePureImport(content) {
  return content.replace(pureImportRE, "");
}
const setupFunctionRE = /function setup\([\s\S]+\)\s+?\{[\s\S]+return __returned__\n\}/;
function transferSetupPosition(content) {
  const match = content.match(setupFunctionRE);
  if (match) {
    const setupFunction = match[0];
    return content.replace(setupFunction, "").replace("setup})", setupFunction.slice("function ".length) + "\n\r})");
  }
  return content;
}

const exportDefaultRE = /export\s+default/;
const exportDefaultClassRE = /(?:(?:^|\n|;)\s*)export\s+default\s+class\s+([\w$]+)/;
const noScriptContent = "import { defineComponent } from 'vue'\nexport default defineComponent({})";
const _require = node_module.createRequire((typeof document === 'undefined' ? new (require('u' + 'rl').URL)('file:' + __filename).href : (document.currentScript && document.currentScript.src || new URL('index.cjs', document.baseURI).href)));
let index = 1;
let compileRoot = null;
let compiler;
let vue;
function requireCompiler() {
  if (!compiler) {
    if (compileRoot) {
      try {
        compiler = _require(_require.resolve("vue/compiler-sfc", { paths: [compileRoot] }));
      } catch (e) {
        try {
          compiler = _require(_require.resolve("@vue/compiler-sfc", { paths: [compileRoot] }));
        } catch (e2) {
        }
      }
    }
    if (!compiler) {
      try {
        compiler = _require("vue/compiler-sfc");
      } catch (e) {
        try {
          compiler = _require("@vue/compiler-sfc");
        } catch (e2) {
          throw new Error("@vue/compiler-sfc is not present in the dependency tree.\n");
        }
      }
    }
  }
  return compiler;
}
function isVue3() {
  if (!vue) {
    if (compileRoot) {
      try {
        vue = _require(_require.resolve("vue", { paths: [compileRoot] }));
      } catch (e) {
      }
    }
    if (!vue) {
      try {
        vue = _require("vue");
      } catch (e) {
        throw new Error("vue is not present in the dependency tree.\n");
      }
    }
  }
  return vue.version.startsWith("3");
}
function setCompileRoot(root) {
  if (root && root !== compileRoot) {
    compileRoot = root;
    compiler = null;
  }
}
function parseCode(code) {
  const { parse } = requireCompiler();
  let descriptor;
  if (isVue3()) {
    descriptor = parse(code).descriptor;
  } else {
    descriptor = parse({ source: code });
  }
  return descriptor;
}
function compileVueCode(code) {
  const { compileScript, rewriteDefault } = requireCompiler();
  const descriptor = parseCode(code);
  const { script, scriptSetup } = descriptor;
  let content = null;
  let ext = null;
  if (script || scriptSetup) {
    if (scriptSetup) {
      const compiled = compileScript(descriptor, {
        id: `${index++}`
      });
      const classMatch = compiled.content.match(exportDefaultClassRE);
      const plugins = scriptSetup.lang === "ts" ? ["typescript", "decorators-legacy"] : void 0;
      if (classMatch) {
        content = compiled.content.replace(exportDefaultClassRE, "\nclass $1") + `
const _sfc_main = ${classMatch[1]}`;
        if (exportDefaultRE.test(content)) {
          content = rewriteDefault(compiled.content, "_sfc_main", plugins);
        }
      } else {
        content = rewriteDefault(compiled.content, "_sfc_main", plugins);
      }
      content = transferSetupPosition(content);
      content = content.replace(/(const __returned__\s?=\s?\{[\s\S]+?)(props)(\s?\})/, "$1props: props as any$3").replace(
        /(const __returned__\s?=\s?\{[\s\S]+?)(props,)([\s\S]+?)/,
        "$1props: props as any,$3"
      );
      content += "\nexport default _sfc_main\n";
      ext = scriptSetup.lang || "js";
    } else if (script && script.content) {
      content = rewriteDefault(
        script.content,
        "_sfc_main",
        script.lang === "ts" ? ["typescript", "decorators-legacy"] : void 0
      );
      content += "\nexport default _sfc_main\n";
      ext = script.lang || "js";
    }
  } else {
    content = noScriptContent;
    ext = "ts";
  }
  return { content, ext };
}

const dtsRE$1 = /\.d\.tsx?$/;
function rollupDeclarationFiles({
  root,
  tsConfigPath,
  outputDir,
  entryPath,
  fileName,
  compilerOptions
}) {
  const configObjectFullPath = node_path.resolve(root, "api-extractor.json");
  const packageJsonLookup = new nodeCoreLibrary.PackageJsonLookup();
  const packageJsonFullPath = packageJsonLookup.tryGetPackageJsonFilePathFor(configObjectFullPath);
  if (!dtsRE$1.test(fileName)) {
    fileName += ".d.ts";
  }
  const extractorConfig = apiExtractor.ExtractorConfig.prepare({
    configObject: {
      projectFolder: root,
      mainEntryPointFilePath: entryPath,
      compiler: {
        tsconfigFilePath: tsConfigPath,
        overrideTsconfig: compilerOptions
      },
      apiReport: {
        enabled: false,
        reportFileName: "<unscopedPackageName>.api.md"
      },
      docModel: {
        enabled: false
      },
      dtsRollup: {
        enabled: true,
        publicTrimmedFilePath: node_path.resolve(outputDir, fileName)
      },
      tsdocMetadata: {
        enabled: false
      },
      messages: {
        compilerMessageReporting: {
          default: {
            logLevel: "none"
          }
        },
        extractorMessageReporting: {
          default: {
            logLevel: "none"
          }
        }
      }
    },
    configObjectFullPath,
    packageJsonFullPath
  });
  const compilerState = apiExtractor.CompilerState.create(extractorConfig, {
    localBuild: true,
    showVerboseMessages: false
  });
  const sourceMapper = new SourceMapper_js.SourceMapper();
  const messageRouter = new MessageRouter_js.MessageRouter({
    workingPackageFolder: root,
    messageCallback: void 0,
    messagesConfig: extractorConfig.messages,
    showVerboseMessages: false,
    showDiagnostics: false,
    tsdocConfiguration: extractorConfig.tsdocConfiguration,
    sourceMapper
  });
  const collector = new Collector_js.Collector({
    program: compilerState.program,
    messageRouter,
    extractorConfig,
    sourceMapper
  });
  collector.analyze();
  DtsRollupGenerator_js.DtsRollupGenerator.writeTypingsFile(
    collector,
    extractorConfig.publicTrimmedFilePath,
    DtsRollupGenerator_js.DtsRollupKind.PublicRelease,
    extractorConfig.newlineKind
  );
}

const noneExport = "export {};\n";
const vueRE = /\.vue$/;
const tsRE = /\.(m|c)?tsx?$/;
const jsRE = /\.(m|c)?jsx?$/;
const dtsRE = /\.d\.tsx?$/;
const tjsRE = /\.(m|c)?(t|j)sx?$/;
const watchExtensionRE = /\.(vue|(m|c)?(t|j)sx?)$/;
const fullRelativeRE = /^\.\.?\//;
const defaultIndex = "index.d.ts";
const noop = () => {
};
const logPrefix = kolorist.cyan("[vite:dts]");
const bundleDebug = debug("vite-plugin-dts:bundle");
function dtsPlugin(options = {}) {
  const {
    tsConfigFilePath = "tsconfig.json",
    aliasesExclude = [],
    cleanVueFileName = false,
    staticImport = false,
    clearPureImport = true,
    insertTypesEntry = false,
    rollupTypes = false,
    noEmitOnError = false,
    skipDiagnostics = false,
    logDiagnostics = void 0,
    copyDtsFiles = true,
    libFolderPath = void 0,
    afterDiagnostic = noop,
    beforeWriteFile = noop,
    afterBuild = noop
  } = options;
  const compilerOptions = options.compilerOptions ?? {};
  let root;
  let entryRoot = options.entryRoot ?? "";
  let libName;
  let indexName;
  let aliases;
  let entries;
  let logger;
  let project;
  let tsConfigPath;
  let outputDirs;
  let isBundle = false;
  let include;
  let exclude;
  let filter;
  const sourceDtsFiles = /* @__PURE__ */ new Set();
  let hasJsVue = false;
  let allowJs = false;
  return {
    name: "vite:dts",
    apply: "build",
    enforce: "pre",
    config(config) {
      if (isBundle)
        return;
      const aliasOptions = config?.resolve?.alias ?? [];
      if (isNativeObj(aliasOptions)) {
        aliases = Object.entries(aliasOptions).map(([key, value]) => {
          return { find: key, replacement: value };
        });
      } else {
        aliases = ensureArray(aliasOptions);
      }
      if (aliasesExclude.length > 0) {
        aliases = aliases.filter(
          ({ find }) => !aliasesExclude.some(
            (alias) => alias && (isRegExp(find) ? find.toString() === alias.toString() : isRegExp(alias) ? find.match(alias)?.[0] : find === alias)
          )
        );
      }
    },
    configResolved(config) {
      if (isBundle)
        return;
      logger = config.logger;
      if (logDiagnostics != null) {
        logger.warn(
          kolorist.yellow(
            `
${kolorist.cyan(
              "[vite:dts]"
            )} 'logDiagnostics' has been deprecated, the original feature now following 'skipDiagnostics'.
`
          )
        );
      }
      if (!config.build.lib) {
        logger.warn(
          kolorist.yellow(
            `
${kolorist.cyan(
              "[vite:dts]"
            )} You building not a library that may not need to generate declaration files.
`
          )
        );
        libName = "_default";
        indexName = defaultIndex;
      } else {
        const filename = config.build.lib.fileName ?? defaultIndex;
        const entry = typeof config.build.lib.entry === "string" ? config.build.lib.entry : Object.values(config.build.lib.entry)[0];
        libName = config.build.lib.name || "_default";
        indexName = typeof filename === "string" ? filename : filename("es", entry);
        if (!dtsRE.test(indexName)) {
          indexName = `${tjsRE.test(indexName) ? indexName.replace(tjsRE, "") : indexName}.d.ts`;
        }
      }
      root = ensureAbsolute(options.root ?? "", config.root);
      tsConfigPath = ensureAbsolute(tsConfigFilePath, root);
      outputDirs = options.outputDir ? ensureArray(options.outputDir).map((d) => ensureAbsolute(d, root)) : [ensureAbsolute(config.build.outDir, root)];
      if (!outputDirs[0]) {
        logger.error(
          kolorist.red(
            `
${kolorist.cyan(
              "[vite:dts]"
            )} Can not resolve declaration directory, please check your vite config and plugin options.
`
          )
        );
        return;
      }
      setCompileRoot(root);
      compilerOptions.rootDir || (compilerOptions.rootDir = root);
      project = new tsMorph.Project({
        compilerOptions: mergeObjects(compilerOptions, {
          noEmitOnError,
          outDir: ".",
          declarationDir: null,
          noUnusedParameters: false,
          declaration: true,
          noEmit: false,
          emitDeclarationOnly: true
        }),
        tsConfigFilePath: tsConfigPath,
        skipAddingFilesFromTsConfig: true,
        libFolderPath: libFolderPath ? ensureAbsolute(libFolderPath, root) : void 0
      });
      allowJs = project.getCompilerOptions().allowJs ?? false;
      const tsConfig = typescript.readConfigFile(tsConfigPath, project.getFileSystem().readFileSync).config ?? {};
      const parentTsConfigPath = tsConfig.extends && ensureAbsolute(tsConfig.extends, root);
      const parentTsConfig = parentTsConfigPath ? typescript.readConfigFile(parentTsConfigPath, project.getFileSystem().readFileSync).config : {};
      include = ensureArray(
        options.include ?? tsConfig.include ?? parentTsConfig.include ?? "**/*"
      ).map(normalizeGlob);
      exclude = ensureArray(
        options.exclude ?? tsConfig.exclude ?? parentTsConfig.exclude ?? "node_modules/**"
      ).map(normalizeGlob);
      filter = pluginutils.createFilter(include, exclude, { resolve: root });
    },
    buildStart(inputOptions) {
      if (Array.isArray(inputOptions.input)) {
        entries = inputOptions.input.reduce((prev, current) => {
          prev[node_path.basename(current)] = current;
          return prev;
        }, {});
      } else {
        entries = { ...inputOptions.input };
      }
    },
    transform(code, id) {
      if (!filter(id)) {
        return null;
      }
      if (vueRE.test(id)) {
        const { content, ext } = compileVueCode(code);
        if (content) {
          if (ext === "js" || ext === "jsx")
            hasJsVue = true;
          project.createSourceFile(`${id}.${ext || "js"}`, content, { overwrite: true });
        }
      } else if (!id.includes(".vue?vue") && (tsRE.test(id) || allowJs && jsRE.test(id))) {
        project.createSourceFile(id, code, { overwrite: true });
      }
      return null;
    },
    watchChange(id) {
      if (watchExtensionRE.test(id)) {
        isBundle = false;
        if (project) {
          const sourceFile = project.getSourceFile(vite.normalizePath(id));
          sourceFile && project.removeSourceFile(sourceFile);
        }
      }
    },
    async closeBundle() {
      if (!outputDirs || !project || isBundle)
        return;
      logger.info(kolorist.green(`
${logPrefix} Start generate declaration files...`));
      bundleDebug("start");
      isBundle = true;
      sourceDtsFiles.clear();
      const startTime = Date.now();
      const includedFileSet = /* @__PURE__ */ new Set();
      if (include && include.length) {
        const files = await glob(include, {
          cwd: root,
          absolute: true,
          ignore: exclude
        });
        files.forEach((file) => {
          if (dtsRE.test(file)) {
            sourceDtsFiles.add(project.addSourceFileAtPath(file));
            if (!copyDtsFiles) {
              return;
            }
            includedFileSet.add(file);
            return;
          }
          includedFileSet.add(`${tjsRE.test(file) ? file.replace(tjsRE, "") : file}.d.ts`);
        });
        if (hasJsVue) {
          if (!allowJs) {
            logger.warn(
              kolorist.yellow(
                `${kolorist.cyan(
                  "[vite:dts]"
                )} Some js files are referenced, but you may not enable the 'allowJs' option.`
              )
            );
          }
          project.compilerOptions.set({ allowJs: true });
        }
        bundleDebug("collect files");
      }
      project.resolveSourceFileDependencies();
      bundleDebug("resolve");
      if (!skipDiagnostics) {
        const diagnostics = project.getPreEmitDiagnostics();
        if (diagnostics?.length) {
          logger.warn(project.formatDiagnosticsWithColorAndContext(diagnostics));
        }
        if (typeof afterDiagnostic === "function") {
          const result = afterDiagnostic(diagnostics);
          isPromise(result) && await result;
        }
        bundleDebug("diagnostics");
      }
      const dtsOutputFiles = Array.from(sourceDtsFiles).map((sourceFile) => ({
        path: sourceFile.getFilePath(),
        content: sourceFile.getFullText()
      }));
      const service = project.getLanguageService();
      const outputFiles = project.getSourceFiles().map(
        (sourceFile) => service.getEmitOutput(sourceFile, true).getOutputFiles().map((outputFile) => ({
          path: vite.normalizePath(node_path.resolve(root, outputFile.compilerObject.name)),
          content: outputFile.getText()
        }))
      ).flat().concat(dtsOutputFiles);
      bundleDebug("emit");
      if (!entryRoot) {
        entryRoot = queryPublicPath(outputFiles.map((file) => file.path));
      }
      entryRoot = ensureAbsolute(entryRoot, root);
      const wroteFiles = /* @__PURE__ */ new Set();
      const outputDir = outputDirs[0];
      await runParallel(os.cpus().length, outputFiles, async (outputFile) => {
        let filePath = outputFile.path;
        let content = outputFile.content;
        const isMapFile = filePath.endsWith(".map");
        if (!includedFileSet.has(isMapFile ? filePath.slice(0, -4) : filePath) || clearPureImport && content === noneExport) {
          return;
        }
        if (!isMapFile && content && content !== noneExport) {
          content = clearPureImport ? removePureImport(content) : content;
          content = transformAliasImport(filePath, content, aliases, aliasesExclude);
          content = staticImport || rollupTypes ? transformDynamicImport(content) : content;
        }
        filePath = node_path.resolve(
          outputDir,
          node_path.relative(entryRoot, cleanVueFileName ? filePath.replace(".vue.d.ts", ".d.ts") : filePath)
        );
        if (typeof beforeWriteFile === "function") {
          const result = beforeWriteFile(filePath, content);
          if (result === false)
            return;
          if (result && isNativeObj(result)) {
            filePath = result.filePath ?? filePath;
            content = result.content ?? content;
          }
        }
        await fs.mkdir(node_path.dirname(filePath), { recursive: true });
        await fs.writeFile(
          filePath,
          cleanVueFileName ? content.replace(/['"](.+)\.vue['"]/g, '"$1"') : content,
          "utf-8"
        );
        wroteFiles.add(vite.normalizePath(filePath));
      });
      bundleDebug("output");
      if (insertTypesEntry || rollupTypes) {
        const pkgPath = node_path.resolve(root, "package.json");
        const pkg = fs.existsSync(pkgPath) ? JSON.parse(await fs.readFile(pkgPath, "utf-8")) : {};
        const entryNames = Object.keys(entries);
        const types = pkg.types || pkg.typings;
        const multiple = entryNames.length > 1;
        const typesPath = types ? node_path.resolve(root, types) : node_path.resolve(outputDir, indexName);
        for (const name of entryNames) {
          let filePath = multiple ? node_path.resolve(outputDir, name.replace(tsRE, ".d.ts")) : typesPath;
          if (fs.existsSync(filePath))
            continue;
          const index = node_path.resolve(
            outputDir,
            node_path.relative(entryRoot, entries[name].replace(tsRE, ".d.ts"))
          );
          let fromPath = vite.normalizePath(node_path.relative(node_path.dirname(filePath), index));
          fromPath = fromPath.replace(dtsRE, "");
          fromPath = fullRelativeRE.test(fromPath) ? fromPath : `./${fromPath}`;
          let content = `export * from '${fromPath}'
`;
          if (fs.existsSync(index)) {
            const entryCodes = await fs.readFile(index, "utf-8");
            if (entryCodes.includes("export default")) {
              content += `import ${libName} from '${fromPath}'
export default ${libName}
`;
            }
          }
          let result;
          if (typeof beforeWriteFile === "function") {
            result = beforeWriteFile(filePath, content);
            if (result && isNativeObj(result)) {
              filePath = result.filePath ?? filePath;
              content = result.content ?? content;
            }
          }
          if (result !== false) {
            await fs.writeFile(filePath, content, "utf-8");
            wroteFiles.add(vite.normalizePath(filePath));
          }
        }
        bundleDebug("insert index");
        if (rollupTypes) {
          logger.info(kolorist.green(`${logPrefix} Start rollup declaration files...`));
          const rollupFiles = /* @__PURE__ */ new Set();
          if (multiple) {
            for (const name of entryNames) {
              const path = node_path.resolve(outputDir, name.replace(tsRE, ".d.ts"));
              rollupDeclarationFiles({
                root,
                tsConfigPath,
                compilerOptions,
                outputDir,
                entryPath: path,
                fileName: node_path.basename(path)
              });
              const wroteFile = vite.normalizePath(path);
              wroteFiles.delete(wroteFile);
              rollupFiles.add(wroteFile);
            }
          } else {
            rollupDeclarationFiles({
              root,
              tsConfigPath,
              compilerOptions,
              outputDir,
              entryPath: typesPath,
              fileName: node_path.basename(typesPath)
            });
            const wroteFile = vite.normalizePath(typesPath);
            wroteFiles.delete(wroteFile);
            rollupFiles.add(wroteFile);
          }
          await runParallel(os.cpus().length, Array.from(wroteFiles), (f) => fs.unlink(f));
          removeDirIfEmpty(outputDir);
          wroteFiles.clear();
          for (const file of rollupFiles) {
            wroteFiles.add(file);
          }
          if (copyDtsFiles) {
            await runParallel(os.cpus().length, dtsOutputFiles, async ({ path, content }) => {
              const filePath = node_path.resolve(outputDir, node_path.basename(path));
              await fs.writeFile(filePath, content, "utf-8");
              wroteFiles.add(vite.normalizePath(filePath));
            });
          }
          bundleDebug("rollup");
        }
      }
      if (outputDirs.length > 1) {
        const dirs = outputDirs.slice(1);
        await runParallel(os.cpus().length, Array.from(wroteFiles), async (wroteFile) => {
          const relativePath = node_path.relative(outputDir, wroteFile);
          const content = await fs.readFile(wroteFile, "utf-8");
          await Promise.all(
            dirs.map(async (dir) => {
              const filePath = node_path.resolve(dir, relativePath);
              await fs.mkdir(node_path.dirname(filePath), { recursive: true });
              await fs.writeFile(filePath, content, "utf-8");
            })
          );
        });
      }
      if (typeof afterBuild === "function") {
        const result = afterBuild();
        isPromise(result) && await result;
      }
      bundleDebug("finish");
      logger.info(kolorist.green(`${logPrefix} Declaration files built in ${Date.now() - startTime}ms.
`));
    }
  };
}

module.exports = dtsPlugin;
